from .utils import compare_two_camel_case_words
# from utils import compare_two_camel_case_words
class ResourceNodeRegistry:
    _registry = {}

    def __init_subclass__(cls, service_name=None, name=None, **kwargs):
        super().__init_subclass__(**kwargs)
        # TODO: check if service_name is correct get_available_service_names
        service_dict = cls._registry.get(service_name, False)
        if not service_dict:
            cls._registry[service_name] = {name: cls}
        else:
            custom_resource_node = service_dict.get(name, False)
            if custom_resource_node:
                raise Exception(
                    f"A custom ResourceNode is already registered with Service: {service_name}, Resource Node: {name}. Duplication is not allowed.")
            else:
                cls._registry[service_name][name] = cls

    def _search_registry_for_service(cls, service_name):
        return cls._registry.get(service_name, {})

    def find_custom_class_for_resource_node(cls, service_name, node_name):
        services_custom_nodes = cls._search_registry_for_service(service_name)
        for custom_cls_name, _cls_obj in services_custom_nodes.items():
            # if compare_two_camel_case_words(custom_cls_name, node_name):
            if custom_cls_name.lower() == node_name.lower():
                return _cls_obj
        return False


from .nodes import ResourceNode
from .registries import ResourceNodeRegistry
from .logs import get_logger
from .relations import FindRelationResultTypes
# from nodes import ResourceNode
# from registries import ResourceNodeRegistry
# from logs import get_logger
# from relations import FindRelationResultTypes


logger = get_logger(__name__)

# class EC2_Vpc(ResourceNode, ResourceNodeRegistry, service_name="ec2", name="Vpc"):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         # print("VPC init")


# class EC2_Instance(ResourceNode, ResourceNodeRegistry, service_name="ec2", name="Instance"):
#     def get_parameter_aliases_of_operation(self, operation_name):
#         result = super().get_parameter_aliases_of_operation(operation_name)
#         return result
        # print("Instance init")

class EC2_SecurityGroups(ResourceNode, ResourceNodeRegistry, service_name="ec2", name="SecurityGroups"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def find_best_relations_for_operation(self, operation_name, relation_map):
        r = super().find_best_relations_for_operation(operation_name, relation_map)
        
"""
1. check for already existing operation data
2. get relations of the operation
3. find the dependent operations to call
4. read the target dependent operation
5. generate JMESpath selectors with the relations
6. search with JMESpath to get raw_parameters
7. generate real api parameters from raw_parameters
8. read the current operation
"""
class S3_BucketLifecycleConfiguration(ResourceNode, ResourceNodeRegistry, service_name="s3", name="BucketLifecycleConfiguration"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "search_shape_name": "Bucket",
            "target_shape_name": "Name",
            "target_shape_type": "string",
            "operation_name": "ListBuckets",
            "target_path": "Buckets[*].Name"
        }]

    # def find_best_relations_for_operation(self, operation_name, relation_map):
    #     r = super().find_best_relations_for_operation(operation_name, relation_map)
    #     return r
    
    def _generate_jmespath_selector_from_relations(self, operation_name, relation_list):
        r = super()._generate_jmespath_selector_from_relations(operation_name, relation_list)
        return r
    
    def create_valid_api_parameters_list(self, operation_name, raw_api_parameters_list, relations_of_operation):
        r = super().create_valid_api_parameters_list(operation_name, raw_api_parameters_list, relations_of_operation)
        return r
    
class IAM_Policy(ResourceNode, ResourceNodeRegistry, service_name="iam", name="Policy"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "search_shape_name": "PolicyArn",
            "target_shape_name": "Arn",
            "target_shape_type": "string",
            "operation_name": "ListPolicies",
            "target_path": "Policies[*].Arn"
        }]

class IAM_User(ResourceNode, ResourceNodeRegistry, service_name="iam", name="User"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "search_shape_name": "UserName",
            "target_shape_name": "UserName",
            "target_shape_type": "string",
            "operation_name": "ListUsers",
            "target_path": "Users[*].UserName"
        }]

class IAM_Role(ResourceNode, ResourceNodeRegistry, service_name="iam", name="Role"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "search_shape_name": "RoleName",
            "target_shape_name": "RoleName",
            "target_shape_type": "string",
            "operation_name": "ListRoles",
            "target_path": "Roles[*].RoleName"
        }]
    
class IAM_AccessKeys(ResourceNode, ResourceNodeRegistry, service_name="iam", name="AccessKeys"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "search_shape_name": "AccessKeyId",
            "target_shape_name": "AccessKeyId",
            "target_shape_type": "string",
            "operation_name": "ListAccessKeys",
            "target_path": "AccessKeyMetadata[*].AccessKeyId"
        }]
        
class ECS_Cluster(ResourceNode, ResourceNodeRegistry, service_name="ecs", name="Clusters"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def _generate_jmespath_selector_from_relations(self, operation_name, relation_list):
        r = super()._generate_jmespath_selector_from_relations(operation_name, relation_list)
        if operation_name == 'ListClusters':
            return '[*].{cluster: clusterArns}[][]'
        return r
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "search_shape_name": "cluster",
            "target_shape_name": "cluster",
            "target_shape_type": "string",
            "operation_name": "ListClusters",
            "target_path": "clusterArns"
        }]
class ECS_Services(ResourceNode, ResourceNodeRegistry, service_name="ecs", name="Services"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_required_parameter_names_from_operation_name(self, operation_name):
        r = super().get_required_parameter_names_from_operation_name(operation_name)
        if operation_name == 'ListServices':
            return ['cluster']
        return r

    def find_best_relations_for_operation(self, operation_name, relation_map):
        r = super().find_best_relations_for_operation(operation_name, relation_map)
        if operation_name == 'DescribeServices':
            return [{
                "search_shape_name": "services",
                "target_shape_name": "serviceArns",
                "target_shape_type": "string",
                "operation_name": "ListServices",
                "target_path": "serviceArns"
            },
            {
                "search_shape_name": "cluster",
                "target_shape_name": "cluster",
                "target_shape_type": "string",
                "operation_name": "ListServices",
                "target_path": "__args__.cluster"
            }
            ]
        return r
    
    # def define_extra_relations(self):
    #     r= super().define_extra_relations()
   

class CodeBuild_BuildsForProject(ResourceNode, ResourceNodeRegistry, service_name="codebuild", name="BuildsForProject"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _generate_jmespath_selector_from_relations(self, operation_name, relation_list):
        r = super()._generate_jmespath_selector_from_relations(operation_name, relation_list)
        if operation_name == 'ListBuildsForProject':
            return '[].projects'
        return r

    def find_best_relations_for_operation(self, operation_name, relation_map):
        r = super().find_best_relations_for_operation(operation_name, relation_map)
        if operation_name == 'ListBuildsForProject':
            return [{
                "search_shape_name": "projectName",
                "target_shape_name": "projectName",
                "target_shape_type": "list",
                "operation_name": "ListProjects",
                "target_path": "[].projects"
            }
            ], FindRelationResultTypes.RelationsFound
        return r
        
class Lambda_Function(ResourceNode, ResourceNodeRegistry, service_name="lambda", name="Function"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "search_shape_name": "FunctionName",
            "target_shape_name": "FunctionName",
            "target_shape_type": "string",
            "operation_name": "ListFunctions",
            "target_path": "Functions[*].FunctionName"
        }]


from .utils import camel_case_split, ifind_similar_names_in_list , compare_two_token_lists, str_relations
from .botocore_utils import *
from .relations import RelationMap, FindRelationResultTypes
from .reader import ServiceReader
from .registries import ResourceNodeRegistry
from .logs import get_logger, get_rich_console
# from utils import camel_case_split, ifind_similar_names_in_list , compare_two_token_lists, str_relations
# from botocore_utils import *
# from relations import RelationMap, FindRelationResultTypes
# from reader import ServiceReader
# from registries import ResourceNodeRegistry
# from logs import get_logger, get_rich_console




from pprint import pprint
import json
import boto3
from itertools import product as cartesian_product
from enum import Enum
import copy
from termcolor import colored
from botocore.utils import ArgumentGenerator
from botocore.validate import validate_parameters
from rich.text import Text
from rich.panel import Panel
from rich.console import Group
from rich.layout import Layout

logger = get_logger(__name__)


_resource_node_registry = ResourceNodeRegistry()
argument_generator = ArgumentGenerator()
console = get_rich_console()

class ResourceNode:
    def __init__(self, service_node: 'ServiceNode', name: str, operation_names):
        self.service_node = service_node
        self.name = name
        self.operation_names = operation_names
        self._operation_models = {}
        
    def define_extra_relations(self):
        return []
        
    def _generate_jmespath_selector_from_relations(self, operation_name, relation_list):
        _for_all_the_responses = '[*].'
        _flatten_two_times = '[][]'
        if not relation_list:
            return False
        # [*].Stacks[*].{StackName: StackName, age: age}[] -> [*].Stacks[*].StackName[][]
        # [*].Stacks[*].StackName[], [*].Stacks[*].Description[]  
        target_shape_and_min_nested_target_paths = None
        if len(relation_list) == 1:
            relation = relation_list[0]
            selected_target_path = min(relation.get('target_path').split(','), key=lambda tp: tp.count('.'))
            target_shape_and_min_nested_target_paths = [
                (relation.get('target_shape_name'), selected_target_path)
            ]
        else:
            # multiple relations
            # all relations point to same operation_name
            target_shape_and_min_nested_target_paths = [    
                (
                    relation.get('target_shape_name'), 
                    min(
                        relation.get('target_path').split(','), 
                        key=lambda tp: tp.count('.')
                    )
                )
                for relation in relation_list
            ]
            # all_selected_target_paths_same_depth = False
            # if len(relations_and_min_nested_target_paths) >= 2:

            # ensure target paths are same deep nested
            first_target_shape_and_target_path, *rest_of_target_shape_and_target_path = target_shape_and_min_nested_target_paths
            _, first_target_path = first_target_shape_and_target_path
            first_target_path_deepness = first_target_path.count('.')
            first_target_path_before_last_attribute = first_target_path.rsplit('.', maxsplit=1)[0]
            all_selected_target_paths_same_depth = all([
                r_target_path.count('.') == first_target_path_deepness \
                    and r_target_path.rsplit('.', maxsplit=1)[0] == first_target_path_before_last_attribute
                for _, r_target_path in rest_of_target_shape_and_target_path
            ])
                    
            if not all_selected_target_paths_same_depth:
                logger.debug(f"[red]DOES NOT START WITH same json nest: {target_shape_and_min_nested_target_paths}")
                return False
                
        jmespath_nested_selector = None
        
        # if len(target_shape_and_min_nested_target_paths) == 1:
        #     jmespath_nested_selector
        # else:
        _jmespath_selectors_list = []
        before_last_attr = None
        for target_shape, target_path in target_shape_and_min_nested_target_paths:
            if '.' not in target_path:
                target_path = f".{target_path}"
            before_last_attr, last_attribute = target_path.rsplit('.', maxsplit=1)
                
            jmespath_nested_selector = f"{target_shape}: {last_attribute}"
            _jmespath_selectors_list.append(jmespath_nested_selector)
        if not before_last_attr:
             before_last_attr = ''
        jmespath_nested_selector = before_last_attr+'.{'+', '.join(_jmespath_selectors_list)+'}'
        jmespath_nested_selector
        flattened_jmespath_nested_selector = f"{_for_all_the_responses}{jmespath_nested_selector}{_flatten_two_times}"
        return flattened_jmespath_nested_selector
    
    def __________(self):
        # get jmes path, 
        pass
    
    def create_valid_api_parameters_list(self, operation_name, raw_api_parameters_list, relations_of_operation):
        operation_model = self.get_operation_model(operation_name)
        required_parameter_names = self.get_required_parameter_names_from_operation_name(operation_name)

        input_shape = get_input_shape(operation_model)
        generated = argument_generator.generate_skeleton(input_shape)
        
        result = {}

        for r_parameter_name in required_parameter_names:
            r_key = find_key_in_dict_keys(r_parameter_name, generated)
            if r_key:
                result[r_key] = generated[r_key]
                
        # handle MaxResults
        max_results_value = get_max_results_value_from_shape(input_shape)
        max_results_key = find_key_in_dict_keys('maxresults', generated)
        if max_results_key and max_results_value:
            result[max_results_key]=max_results_value
        # sometimes MaxResults can be seen as non required parameter, but in fact is
        if not required_parameter_names:
            return result
        
        
        search_to_target_names = {relation.get('target_shape_name'):relation.get('search_shape_name') for relation in relations_of_operation}
        
        generated_api_params = []
        for raw_api_parameter_dict in raw_api_parameters_list:
            result_copy = copy.deepcopy(result)
            for raw_key in raw_api_parameter_dict.keys():
                found_parameter_key = find_key_in_dict_keys(raw_key, result_copy)
                if not found_parameter_key:
                    # search to target name
                    new_key = find_key_in_dict_keys(raw_key, search_to_target_names)
                    if new_key:
                        found_parameter_key = find_key_in_dict_keys(search_to_target_names[new_key], result_copy)
                result_copy[found_parameter_key] = raw_api_parameter_dict[raw_key]
            generated_api_params.append(result_copy)
        return generated_api_params
        # Handle NextToken and DryRun
        # dry_run_key_name = find_key_in_dict_keys('dryrun', generated)
        # generated.pop(dry_run_key_name)
        # next_token_key_name = find_key_in_dict_keys('nexttoken', generated)
        # generated.pop(next_token_key_name)
        # generated
        # # handle multiple required parameters
        # generated_api_params_list = []
        # for r_parameter_name in required_parameter_names:
        #     r_parameter_name_key = find_key_in_dict_keys(r_parameter_name, generated)
        #     r_parameter_value = generated.get(r_parameter_name_key)
        #     # if type
            
        #     # result[r_parameter_name_key] = ?
        # generated

    
    def print_operation(self, operation_name):
        operation_panel = self._rich_operation_details_panel(operation_name)
        console.print(operation_panel)

    # TODO: print operation and resource node, by rich methods 
    def _rich_operation_details_panel(self, operation_name):
        
        operation_model = self.get_operation_model(operation_name)
        input_shape = get_input_shape(operation_model)
        output_shape = operation_model.output_shape
        
        input_shape_tree = Text("[yellow]No Input Shape Found")
        if input_shape:
            input_shape_tree = generate_rich_tree_from_shape(input_shape)
        output_shape_tree = generate_rich_tree_from_shape(output_shape)
        layout = Layout()
        layout.split_row(
            Panel(input_shape_tree, title='[yellow]Input Shape'),# subtitle=get_shape_name(input_shape)),
            Panel(output_shape_tree, title='[yellow]Output Shape'),# subtitle=get_shape_name(output_shape)),
        )
      
        required_parameters = self.get_required_parameter_names_from_operation_name(operation_name)
        _title = f"Operation: [green][bold]{operation_name}[/]"
        if required_parameters:
            _title = f"{_title}   [white][Required: {', '.join(required_parameters)}]"
        operation_panel = Panel(layout, title=_title, highlight=True, title_align='left')
        return operation_panel

    def find_best_relations_for_operation(self, operation_name, relation_map):
        resource_node = self
        required_parameter_names = resource_node.get_required_parameter_names_from_operation_name(operation_name)
        required_parameter_names_to_relations_map = {
            r_param_name:None 
            for r_param_name in required_parameter_names
        }

        if False:
            # TODO: return false on required parameter and no relation found
            return False, FindRelationResultTypes.NoGeneratedParameters
        
        if not required_parameter_names:
            # no required parameters
            logger.debug(f"NO REQUIRED PARAMETERS. {self.service_node.name}.{operation_name} has no required parameters")
            return True, FindRelationResultTypes.NoRequiredParameters

        if len(required_parameter_names) == 1:
            # only one parameter exists
            selected_relations = None
            single_parameter_name = required_parameter_names[0]
            generated_relations_for_parameter = relation_map.get_parameters_generated_relations(single_parameter_name, operation_name)
            if not generated_relations_for_parameter:
                logger.debug(f"NO RELATIONS GENERATED. {self.service_node.name}.{operation_name} has req parameter: {single_parameter_name}. {generated_relations_for_parameter=}")
                return False, FindRelationResultTypes.NoGeneratedParameters

            selected_relations = resource_node.find_best_relation_for_parameter(single_parameter_name, generated_relations_for_parameter)
            # required_parameter_names_to_relations_map[single_parameter_name]=selected_relations
            if selected_relations:
                logger.debug(f"SINGLE PARAMETER FOUND. {self.service_node.name}.{operation_name} has req parameter: {single_parameter_name}. Relation found: {str_relations(selected_relations)}")
                return selected_relations, FindRelationResultTypes.RelationsFound
            else:
                logger.debug(f"CAN'T DECIDE BTWN RELATIONS. {self.service_node.name}.{operation_name} has req parameters: {required_parameter_names}.")
                return False, FindRelationResultTypes.CantDecideBetweenGeneratedParameters
        else:
            # multiple required parameters, zip and check
            # required_parameter_names = list(required_parameter_names_to_relations_map.keys())
            # temporary_parameters_to_relations_map = dict()
            for r_parameter_name in required_parameter_names:
                generated_relations_for_parameter = relation_map.get_parameters_generated_relations(r_parameter_name, operation_name)
                if generated_relations_for_parameter:
                    required_parameter_names_to_relations_map[r_parameter_name]=generated_relations_for_parameter
                # return False, FindRelationResultTypes.NoGeneratedParameters
                # selected_relations = resource_node.find_best_relation_for_parameter(r_parameter_name, generated_relations_for_parameter)
                # ignore relations to self

            _all_parameters_have_relations = [
                bool(_found_relations) 
                for _found_relations in required_parameter_names_to_relations_map.values()
                
            ]

            if not all(_all_parameters_have_relations):
                # all parameters doesnt have relations
                if any(_all_parameters_have_relations):
                    # some found but not all
                    partial_relations_str = ", ".join([
                        str_relations(_found_relations)
                        for _found_relations in required_parameter_names_to_relations_map.values()
                        if _found_relations
                    ])
                    logger.debug(f"NOT ALL RELATIONS FOUND. {self.service_node.name}.{operation_name} has req parameter: {required_parameter_names}. Partially found relations: {partial_relations_str}")
                    return False, FindRelationResultTypes.SomeRelationsFoundButNotAll
                else:
                    # nothing found
                    logger.debug(f"NO RELATIONS FOUND. {self.service_node.name}.{operation_name} has req parameter: {required_parameter_names}.")
                    return False, FindRelationResultTypes.NoRelations
            
            def check(list_of_relations):
                if not list_of_relations:
                    return False
                first_operation_name = list_of_relations[0].get('operation_name')
                return all([l_o_r.get('operation_name')==first_operation_name for l_o_r in list_of_relations])
       
            possible_relation_combinations = None
            relation_matrix = list(required_parameter_names_to_relations_map.values())
            if relation_matrix:
                relations_cartesian_product = cartesian_product(*relation_matrix)
                possible_relation_combinations = list(filter(check, relations_cartesian_product))            
            
            if not possible_relation_combinations:
                logger.debug(f"CAN'T DECIDE BTWN RELATIONS. {self.service_node.name}.{operation_name} has req parameters: {required_parameter_names}.")
                return False, FindRelationResultTypes.CantDecideBetweenGeneratedParameters
                # TODO: add required parameters to attributes + making them available as relations. add metadata.
            
            # find common target_operation relations across the permutation.
            relation_chosen_from_possible_combinations = self.select_between_possible_relation_combinations_matrix(possible_relation_combinations)    
            if relation_chosen_from_possible_combinations:
                logger.debug(f"MULTIPLE PARAMETERS FOUND. {self.service_node.name}.{operation_name} has req parameters: {required_parameter_names}. Relations found: {str_relations(relation_chosen_from_possible_combinations)}")
                return relation_chosen_from_possible_combinations, FindRelationResultTypes.RelationsFound
            else:
                logger.debug(f"CAN'T DECIDE BTWN RELATIONS. {self.service_node.name}.{operation_name} has req parameters: {required_parameter_names}.")
                return False, FindRelationResultTypes.CantDecideBetweenGeneratedParameters


    def select_between_possible_relation_combinations_matrix(self, possible_relation_combinations):
        if len(possible_relation_combinations)==1:
            return possible_relation_combinations[0]
        # each combination will have the same operation name
        # prefer using non-GET function
        # prefer the operation that has no required parameters
        combination_index_to_score_map = {str(i):0 for i in range(len(possible_relation_combinations))}
        
        for i, relation_combination in enumerate(possible_relation_combinations):
            current_combination_point = 0
            combinations_operation_name = relation_combination[0].get('operation_name') # promised to all have same operation
            
            # negative point if it has required parameters
            combinations_required_parameters = self.get_required_parameter_names_from_operation_name(combinations_operation_name)
            current_combination_point -= len(combinations_required_parameters)
            
            # negative point if its a get function, we'd like to prefer list or describe functions
            if combinations_operation_name.lower().startswith('get'):
                current_combination_point -= 1
            
            combination_index_to_score_map[str(i)] = current_combination_point
        selected_index_str, selected_score = max(combination_index_to_score_map.items(), key=lambda index_and_score: index_and_score[1])
        selected_combination = possible_relation_combinations[int(selected_index_str)]
        
        return selected_combination

    def select_between_possible_relation_combinations_list(self, possible_relation_list):
        if len(possible_relation_list)<1:
            return False
        if len(possible_relation_list)==1:
            return possible_relation_list[0]
        # each combination will have the same operation name
        # prefer using non-GET function
        # prefer the operation that has no required parameters
        combination_index_to_score_map = {str(i):0 for i in range(len(possible_relation_list))}
        
        get_operation_count  = [p_rel.get('operation_name').startswith('Get') for p_rel in possible_relation_list].count(True)
        for i, relation_dict in enumerate(possible_relation_list):
            current_combination_point = 0
            relations_operation_name = relation_dict.get('operation_name') # promised to all have same operation
            
            # negative point if it has required parameters
            combinations_required_parameters = self.get_required_parameter_names_from_operation_name(relations_operation_name)
            current_combination_point -= len(combinations_required_parameters)
            
            # negative point if its a get function, we'd like to prefer list or describe functions
            if get_operation_count <= 1 and relations_operation_name.lower().startswith('get'):
                current_combination_point -= 1
            
            combination_index_to_score_map[str(i)] = current_combination_point
        selected_index_str, selected_score = max(combination_index_to_score_map.items(), key=lambda index_and_score: index_and_score[1])
        selected_combination = possible_relation_list[int(selected_index_str)]
        
        return selected_combination
    
    
    def find_best_relation_for_parameter(self, parameter_name, relation_list):
        _parameter_name_tokens = camel_case_split(parameter_name)
        non_id_parameter_tokens = [
            p_token for p_token in _parameter_name_tokens
            if p_token.lower() not in IDENTIFIER_NAMES
        ]
        
        # if not non_id_parameter_tokens:
        #     # only contains id 
        #     non_id_parameter_tokens  =_parameter_name_tokens

        same_resource_name_relations = []
        for relation_dict in relation_list:
            operation_name = relation_dict.get('operation_name')
            operation_verb, *operation_name_tokens = camel_case_split(operation_name)
            # relation_dict.update({'operation_verb': operation_verb, 'operation_name_tokens': operation_name_tokens})
            # TODO: here
            all_tokens_match = compare_two_token_lists(
                non_id_parameter_tokens, operation_name_tokens)
            if all_tokens_match:
                # merge same operations different target paths
                target_concat_flag = False        
                for i, current_ in enumerate(same_resource_name_relations):
                    # if current_.get('alias', False):
                    #     continue
                    cur_op_name, rel_op_name = current_.get('operation_name'), relation_dict.get('operation_name')
                    cur_target_path, rel_target_path = current_.get('target_path'), relation_dict.get('target_path')
                    if cur_op_name==rel_op_name and rel_target_path not in cur_target_path:
                        current_['target_path'] = ','.join([cur_target_path, rel_target_path])
                        same_resource_name_relations[i]=current_
                        target_concat_flag = True
                        break
                if not target_concat_flag:
                    same_resource_name_relations.append(relation_dict)
                    
        # what we are left with is different operations names 
        selected_relation = self.select_between_possible_relation_combinations_list(same_resource_name_relations)
        if selected_relation:
            return [selected_relation]
        return False
    def get_required_parameter_names_from_operation_name(self, operation_name):
        
        operation_model = self.get_operation_model(operation_name)
        input_shape = get_input_shape(operation_model)
        if not input_shape:
            return []
        required_parameter_names = getattr(input_shape, 'required_members', False)
        max_results_key = find_key_in_dict_keys('maxresults', required_parameter_names)
        if max_results_key:
            required_parameter_names.remove(max_results_key)
        return required_parameter_names
    
    def get_all_required_parameter_names(self):
        all_required_names = []
        for operation_name in self.operation_names:
            required_shapes = self.get_required_parameter_names_from_operation_name(
                operation_name)
            all_required_names.extend(required_shapes)
        return all_required_names
    
    def get_required_parameter_shapes_by_operation_name(self, operation_name):
        operation_model = self.get_operation_model(operation_name)
        required_shapes = get_required_parameter_shapes_from_operation_model(
            operation_model)
        return required_shapes
    
    

    def get_all_required_parameter_shapes(self):
        all_required_shapes = []
        for operation_name in self.operation_names:
            required_shapes = self.get_required_parameter_shapes_by_operation_name(
                operation_name)
            all_required_shapes.extend(required_shapes)
        return all_required_shapes

    def get_operation_model(self, operation_name):
        if operation_name in self._operation_models:
            return self._operation_models[operation_name]
        service_model = self.service_node.get_service_model()
        operation_model = service_model.operation_model(operation_name)
        self._operation_models[operation_name] = operation_model
        return operation_model

    def json(self):
        return {
            "service_node_name": self.service_node.name,
            "name": self.name,
            "operation_names": self.operation_names
        }

    def __str__(self):
        # return f"[{self.service_node.name}.{self.name}]"
        return f"[{self.name}]"


class ServiceNode:
    def __init__(self, name, session):
        self.name = name
        self.session = session
        self.client = self.session.client(self.name)
        self.resource_nodes = None
        self._relation_map = None
        self._read_operation_name_to_tokens_map = None

    def print_resource_node(self, resource_node_name):
        # TODO: check if resource node exists
        operations_panel = self._get_operation_details_panel(resource_node_name)
        if operations_panel:
            console.print(operations_panel)

        
    def _get_operation_details_panel(self, resource_node_name):
        resource_node =  self.get_resource_node_by_name(resource_node_name)
        if not resource_node:
            return False
        panels_for_operations = [
            resource_node._rich_operation_details_panel(operation_name)
            for operation_name in resource_node.operation_names
        ]

        operations_group = Group(
            *panels_for_operations
        )
        return Panel(operations_group, title=f'Resource Node: [blue][bold]{resource_node.name} ', title_align='left')
        
        
        
        
    def get_resource_node_by_name(self, resource_node_name):
        for r_node in self.get_resource_nodes():
            if r_node.name == resource_node_name:
                return r_node
        return False
    
    def json(self):
        return {
            "service_name": self.name,
            # TODO:
        }

    def find_resource_node_by_operation_name(self, operation_name):
        for r_node in self.get_resource_nodes():
            if operation_name in r_node.operation_names:
                return r_node
        return None

    def get_relation_map(self):
        if self._relation_map is not None:
            return self._relation_map
        self._relation_map = RelationMap(self)
        return self._relation_map

    def get_resource_nodes(self):
        if self.resource_nodes is None:
            self.resource_nodes = self._generate_resource_nodes()
        return self.resource_nodes

    def create_resource_node(self, **kwargs):
        service_name = self.name
        resource_node_name = kwargs.get('name', False)
        _ResourceNodeClass = ResourceNode
        if not resource_node_name:
            logger.debug(f"A resource name must be provided on creation.", extra=kwargs)
            raise Exception('ResourceNode name must be provided.')
        _custom_cls_for_resource_node = _resource_node_registry.find_custom_class_for_resource_node(service_name, resource_node_name)
        if _custom_cls_for_resource_node:
            # print(_custom_cls_for_resource_node, 'is used for ', service_name, resource_node_name)
            logger.debug(f"Registry: {_custom_cls_for_resource_node} class is used for  {service_name} {resource_node_name}")
            _ResourceNodeClass = _custom_cls_for_resource_node
        return _ResourceNodeClass(**kwargs)
    
    def _generate_resource_nodes(self):
        generated_resouce_nodes = []
        op_name_to_tokens_map = self.get_read_operation_name_to_tokens_map()
        resource_name_to_operations_map = {}
        for operation_name, op_name_tokens in op_name_to_tokens_map.items():
            verb, *resource_name_list = op_name_tokens
            resource_node_name = ''.join(resource_name_list)
            if resource_node_name not in resource_name_to_operations_map:
                resource_name_to_operations_map[resource_node_name] = [
                    operation_name]
            else:
                resource_name_to_operations_map[resource_node_name].append(
                    operation_name)
                
        for resource_node_name, operations_list in resource_name_to_operations_map.items():
            created_resource_node = self.create_resource_node(
                service_node=self, 
                name=resource_node_name, 
                operation_names=operations_list
            )
            if created_resource_node:
                generated_resouce_nodes.append(created_resource_node)

        if len(generated_resouce_nodes) <= 1:
            return generated_resouce_nodes

        # TODO: decide if this should be kept
        # concat same names
        combined_resource_nodes = []
        _used_for_combining_resource_nodes = []

        for i, gen_resource_node in enumerate(generated_resouce_nodes[:-1]):
            if gen_resource_node in _used_for_combining_resource_nodes:
                continue
            for other_resource_node in generated_resouce_nodes[i+1:]:
                if compare_two_camel_case_words(gen_resource_node.name, other_resource_node.name):
                    # select the shortest name, meaning singular
                    singular_named_resource_node = gen_resource_node
                    plural_named_resource_node = other_resource_node
                    if len(singular_named_resource_node.name) > len(plural_named_resource_node.name):
                        singular_named_resource_node, plural_named_resource_node = plural_named_resource_node, singular_named_resource_node
                    _used_for_combining_resource_nodes.append(
                        plural_named_resource_node)
                    singular_named_resource_node.operation_names.extend(
                        plural_named_resource_node.operation_names)

            if gen_resource_node not in _used_for_combining_resource_nodes:
                combined_resource_nodes.append(gen_resource_node)

        return combined_resource_nodes

    def get_read_operation_name_to_tokens_map(self):
        if self._read_operation_name_to_tokens_map is not None:
            return self._read_operation_name_to_tokens_map
        read_operation_names = self.get_read_operation_names()
        self._read_operation_name_to_tokens_map = self._generate_operation_name_to_tokens_map(
            read_operation_names)
        return self._read_operation_name_to_tokens_map

    def _generate_operation_name_to_tokens_map(self, operation_names=None):
        if operation_names is None:
            operation_names = self.get_operation_names()

        op_name_to_tokens = {
            op_name: camel_case_split(op_name)
            for op_name in operation_names
        }
        return op_name_to_tokens

    def get_read_operation_names(self):
        operation_names = self.get_operation_names()
        read_only_operation_names = [
            op_name
            for op_name in operation_names
            if any([op_name.startswith(read_verb) for read_verb in READ_ONLY_VERBS])
        ]
        return read_only_operation_names

    def get_service_model(self):
        service_model = self.client._service_model
        return service_model

    def get_event_system(self):
        event_system = self.client.meta.events
        return event_system

    def get_operation_names(self):
        service_model = self.get_service_model()
        operation_names = service_model.operation_names
        return operation_names




from .utils import compare_two_camel_case_words, str_relations, ifind_similar_names_in_list
from .botocore_utils import READ_ONLY_VERBS
from .relations import FindRelationResultTypes, SUCCESSFUL_FIND_RELATION_RESULT_TYPES
from .logs import get_logger

# from utils import compare_two_camel_case_words, str_relations, ifind_similar_names_in_list
# from botocore_utils import READ_ONLY_VERBS
# from relations import FindRelationResultTypes, SUCCESSFUL_FIND_RELATION_RESULT_TYPES
# from logs import get_logger

import jmespath
from rich.progress import Progress, track, BarColumn, TextColumn, TaskProgressColumn, TimeRemainingColumn
from rich.table import Column
logger = get_logger(__name__)
class ServiceReader:
    def __init__(self, service_node):
        self.service_node = service_node
        self.response_data = {}

    def add_response_data(self, operation_name, response):
        if operation_name not in self.response_data:
            self.response_data[operation_name] = [response]
        else:
            self.response_data[operation_name].append(response)
            

    def verify_all_required_parameters_in_selected_relations(self, required_parameters, selected_relations):
        if not selected_relations or type(selected_relations) not in (list,tuple):
            return False
        parameter_to_relation_map = {required_parameter_name: None for required_parameter_name in required_parameters}
        for required_parameter_name in parameter_to_relation_map.keys():
            is_parameter_in_relation_map = any([
                compare_two_camel_case_words(required_parameter_name, relation_dict.get('search_shape_name')) 
                 for relation_dict in selected_relations
            ])
            parameter_to_relation_map[required_parameter_name]=is_parameter_in_relation_map
        is_all_parameters_found_in_relations = all(parameter_to_relation_map.values())
        return is_all_parameters_found_in_relations
    
    
    def _call_operation(self, operation_name, api_parameters):
        # TODO: follow next tokens
        client = self.service_node.client
        from botocore.exceptions import ClientError
        try:
            # print(colored(f"{operation_name=} {api_parameters=}",'green'))
            logger.debug(f"CALLING OPERATION: [blue]{operation_name}[/] with parameters: {api_parameters}")
            response = client._make_api_call(operation_name, api_parameters)
            # response = None
            
        except ClientError as e:
            # print(colored(str(e),'red'))
            logger.debug(f"FAILED: CALLING OPERATION. {operation_name} with parameters: {api_parameters}. Exception: {str(e)} ")
            return False
        
        resource_node = self.service_node.find_resource_node_by_operation_name(operation_name)
        if not resource_node:
            return False
        resource_node_name = resource_node.name
        self.add_to_node_data(resource_node_name, operation_name, response)
        response.pop('ResponseMetadata') # FIXME: add back
        response['__args__'] = api_parameters
        return response
    
    
        # if operation_data:
        #     new_api_params = jmespath.search(flattened_jmespath_nested_selector, operation_data)
        # operation_data
        # for 
        # # jmespath.search(f"[*].{target_path}[][]", operation_data)
        
        # target_paths = target_path.split(',')
        # all_found = []
        # for target_path in target_paths:
        #     # found =  
        #     jmespath.search(f"{_for_all_the_responses}{target_path}{_flatten_two_times}", response)
        #     if found:
        #         all_found.extend(found)
        # return all_found

    def _get_dependents_of_single_operation(self, operation_name):
        resource_node = self.service_node.find_resource_node_by_operation_name(operation_name)
        if not resource_node:
            return False
        resource_node_name = resource_node.name
        relations, relation_result_type = self.get_operations_relations(resource_node_name, operation_name)
        success = relation_result_type in SUCCESSFUL_FIND_RELATION_RESULT_TYPES
        if not success:
            logger.debug(f"FAILED to find dependents of the operation: {resource_node_name} -> {operation_name}. Relations result type: {relation_result_type}")
            return False
        
        if relation_result_type in [FindRelationResultTypes.NoRequiredParameters]:
            return True            
        elif relation_result_type in [FindRelationResultTypes.RelationsFound]:
            related_operations = {
                relation.get('operation_name')
                for relation in relations
            }
            return list(related_operations)
        return False
    
    def get_dependent_operations(self, operation_name, _caller_operations_history=False):
        result = [operation_name]
        dependent_operations = self._get_dependents_of_single_operation(operation_name)
        if dependent_operations==False:
            return False
        
        # try to detect self refencing, functions having dependents as: A -> B -> A 
        self_refencing_detected = False
        if type(_caller_operations_history)==list and len(_caller_operations_history)>=2:
            self_refencing_detected = not (len(set(_caller_operations_history))==len(_caller_operations_history))
            
        if self_refencing_detected:
            logger.debug(f"SELF REFERENCING DETECTED. Operation: {operation_name}. {_caller_operations_history=}")
            return False

        if type(dependent_operations)==list:
            if type(_caller_operations_history)==bool and _caller_operations_history==False:
                _caller_operations_history = []
            for d_operation in dependent_operations:

                d_operations_dependencies = self.get_dependent_operations(d_operation, _caller_operations_history=_caller_operations_history+[d_operation])
                if d_operations_dependencies == False:
                    logger.debug(f"FAILED TO FIND NESTED DEPENDENT. Dependent operation: {d_operation}. {_caller_operations_history=}")
                    return False
                if type(d_operations_dependencies)==list:
                    result.extend(d_operations_dependencies)
        elif type(dependent_operations)==boolean and dependent_operations==True:
            # result.append( True) # no more relations
            pass
        
        return result

    def search_node_data(self, resource_node_name, operation_name):
        resource_node_exists = self.response_data.get(resource_node_name, False) != False
        if not resource_node_exists:
            return False
        result = self.response_data[resource_node_name].get(operation_name, False)
        return result 
    
    def add_to_node_data(self, resource_node_name, operation_name, response):
        resource_node_exists = self.response_data.get(resource_node_name, False) != False
        
        if not resource_node_exists:
            self.response_data[resource_node_name] = {}
                
        if operation_name not in self.response_data[resource_node_name]:
            self.response_data[resource_node_name][operation_name] = [response]
        else:
            self.response_data[resource_node_name][operation_name].append(response)

    def read_operation(self, resource_node_name, operation_name, refresh=False):
        # if it has been read called already, return it
        already_existing_data = self.search_node_data(resource_node_name, operation_name)
        if already_existing_data!=False and refresh == False:
            return already_existing_data
        
        resource_node = self.service_node.get_resource_node_by_name(resource_node_name)
        
        if not resource_node:
            logger.debug(f"RESOURCE NODE NOT FOUND. While reading the {resource_node_name}.{operation_name}, Resource Node {resource_node_name} could not be loaded.")
            return False
        
        relations_of_operation, relation_result_type = self.get_operations_relations(resource_node_name, operation_name)
        success = relation_result_type in SUCCESSFUL_FIND_RELATION_RESULT_TYPES
        # req_params = resource_node.get_required_parameter_names_from_operation_name(operation_name)
        if not success:
            # print('\t', colored(operation_name, _color), str(relation_result_type.value), req_params)
            logger.debug(f"FAILED READING OPERATION: {resource_node_name}.{operation_name}. Failed to find relations: {relation_result_type}")
            return False
        
        ########################### NO REQUIRED PARAMETERS #############################################
        if relation_result_type in [FindRelationResultTypes.NoRequiredParameters]:
            relations_of_operation = []
            api_parameters = {}
            raw_api_parameters_list = []
            api_parameters = resource_node.create_valid_api_parameters_list(operation_name, raw_api_parameters_list, relations_of_operation)
            self._call_operation(operation_name, api_parameters)
            return self.search_node_data(resource_node_name, operation_name)
            # return False # FIXME:
        ############################### RELATIONS FOUND #########################################

        elif relation_result_type in [FindRelationResultTypes.RelationsFound]:
            dependent_operations = self.get_dependent_operations(operation_name)
            if dependent_operations == False:
                return False
            self_operation, *other_operations = dependent_operations
            logger.debug(f"OPERATION HAS DEPENDENTS: {self_operation} has dependent operations: {other_operations}. Relations of operation: {str_relations(relations_of_operation)}")
            if other_operations:
                for reverse_other_operation in other_operations[::-1]:
                    logger.debug(f'CALLING DEPENDENT OPERATION: {reverse_other_operation} from {operation_name}')
                    self.read_operation(resource_node_name, reverse_other_operation)
                    if not False:
                        # check if the dependent call is succeeded 
                        pass
                    
            ########################## READ THE TARGET/DEPENDENT OPERATION ##############################################
            if len(relations_of_operation)>=2:
                relations_of_operation

            # get the operation_data
            first_relation, *_ = relations_of_operation
            target_operation_name = first_relation.get('operation_name')
            target_resource_node = self.service_node.find_resource_node_by_operation_name(target_operation_name)
            self.read_operation(target_resource_node.name, target_operation_name)
            operation_data = self.search_node_data(target_resource_node.name, target_operation_name)

            if operation_data==False:
                logger.debug(f"OPERATION DATA NOT FOUND. Tried to read {operation_name}'s dependent [{target_resource_node.name} {target_operation_name}]. Failing.")
                return False
            ########################## Generate Parameters for  ##############################################

            generated_jmespath_nested_selector = target_resource_node._generate_jmespath_selector_from_relations(target_operation_name, relations_of_operation)
            if not generated_jmespath_nested_selector:
                logger.debug(f"CAN'T GENERATE JMESPATH SELECTOR: {target_resource_node.name} {target_operation_name} {generated_jmespath_nested_selector}")
                return False
            logger.debug(f"JMESPATH SELECTOR GENERATED [blue]{generated_jmespath_nested_selector}[/], target operation: [bold][blue][{target_resource_node.name}[/].[green]{target_operation_name}[/]]")

            if not operation_data:
                logger.debug(f"CANT GENERATE API PARAMETERS LIST. NO OPERATION DATA FOUND. {operation_data=}")

            raw_api_parameters_list = jmespath.search(generated_jmespath_nested_selector, operation_data)
            if not raw_api_parameters_list:
                logger.debug(f"CANT GENERATE API PARAMETERS LIST WITH [bold][red]{generated_jmespath_nested_selector}[/], target operation: [bold][blue][{target_resource_node.name}[/].[green]{target_operation_name}[/]], {operation_data=}")
                return False

            api_parameters_list = target_resource_node.create_valid_api_parameters_list(operation_name, raw_api_parameters_list, relations_of_operation)
            api_parameters_list
            
            # FIXME later
            # for api_parameters in track(api_parameters_list, description=f"Calling Operation: {operation_name} with {len(api_parameters_list)} parameters."):
            for api_parameters in api_parameters_list:
                self._call_operation(operation_name, api_parameters)
            operation_data = self.search_node_data(resource_node_name, operation_name)
            return operation_data

            
    def read_resource_node(self, resource_node_name):
        resource_node = self.service_node.get_resource_node_by_name(resource_node_name)
        if not resource_node:
            return False
        # print('\t', colored(resource_node_name,'yellow'))
        # read all operations
        
        for operation_name in resource_node.operation_names:
            # if operation_name != 'GetPasswordData':
            #     continue
            req_params = resource_node.get_required_parameter_names_from_operation_name(operation_name)
            # if req_params:
            req_params
            self.read_operation(resource_node_name, operation_name)
            # print('--------------------------------')
            # print()
          
        return self.response_data.get(resource_node.name)
            # if output == True:
            #     print(' >>>>>>>>>>>>>>>> NO PARAMETERS <<<<<<<<<<<<<<<<<<<<<<')


    def get_operations_relations(self, resource_node_name, operation_name):
        resource_node = self.service_node.get_resource_node_by_name(resource_node_name)
        if not resource_node:
            return False, FindRelationResultTypes.InternalError
        relation_map = self.service_node.get_relation_map()
        required_parameters = resource_node.get_required_parameter_names_from_operation_name(operation_name)

        
        if not required_parameters:
            return True, FindRelationResultTypes.NoRequiredParameters
        else:
            # there are required parameters, try to find every relation for this operation
            relations_of_operation, relation_result_type = resource_node.find_best_relations_for_operation(operation_name, relation_map)
            is_all_parameters_found = self.verify_all_required_parameters_in_selected_relations(required_parameters, relations_of_operation)
            # return relations, relation_result_type
            if relation_result_type not in SUCCESSFUL_FIND_RELATION_RESULT_TYPES:
                # not succeeded
                return False, relation_result_type
            elif not is_all_parameters_found:
                return False, FindRelationResultTypes.SomeRelationsFoundButNotAll
            elif not relations_of_operation:
                return False, FindRelationResultTypes.NoRelations
            else:
                # success
                return relations_of_operation, FindRelationResultTypes.RelationsFound
             

        # if relation_result_type in SUCCESSFUL_FIND_RELATION_RESULT_TYPES:
            # FindRelationResultTypes.SomeRelationsFoundButNotAll
    # def list_available_methods(self):
    #     read_operation_names = self.service_node.get_read_operation_names()
    #     for r_ops_name in read_operation_names:
    #         for read_verb in READ_ONLY_VERBS:
    #             if r_ops_name.startswith(read_verb):

    #     pass



import boto3

class SingletonBaseClass(object):
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(SingletonBaseClass, cls).__new__(cls)
        return cls.instance

class Boto3SessionSingleton(SingletonBaseClass):
    """
    example usage: `singletons.Boto3SessionSingleton().get_session()`

    Extends: SingletonBaseClass
    Args: N/A
    """

    def __init__(self):
        self.session = boto3.session.Session()

    def get_session(self):
        return self.session


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['balcony']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.24.80,<2.0.0',
 'inflect>=6.0.0,<7.0.0',
 'jmespath>=1.0.1,<2.0.0',
 'rich>=12.5.1,<13.0.0',
 'termcolor>=2.0.1,<3.0.0',
 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['balcony = balcony.cli:run_app']}

setup_kwargs = {
    'name': 'arzuhal',
    'version': '0.0.2',
    'description': 'No description',
    'long_description': '# arzuhal',
    'author': 'Oguzhan Yilmaz',
    'author_email': 'oguzhanylmz271@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from functools import lru_cache
from typing import List, Union
from botocore.model import Shape, DenormalizedStructureBuilder, OperationModel

from .utils import compare_two_camel_case_words
from .logs import get_rich_console
# from utils import compare_two_camel_case_words
# from logs import get_rich_console



from collections import namedtuple
from rich.text import Text
from rich.tree import Tree
from rich.layout import Layout
ShapeAndTargetPath = namedtuple('ShapeAndTargetPath', ['shape', 'target_path'])

UNWANTED_SHAPE_NAMES = ('String', 'DateTime', 'Name', 'Id', 'Arn', '__string')
UNWANTED_SHAPE_NAMES_LOWERED = [_.lower() for _ in UNWANTED_SHAPE_NAMES]
SHAPE_SCALAR_TYPES = DenormalizedStructureBuilder.SCALAR_TYPES
SHAPE_COLLECTION_TYPES = ('structure', 'list','map')
READ_ONLY_VERBS = ('Describe', 'List', 'Get')
IDENTIFIER_NAMES = ('arn', 'id', 'name', 'arns', 'ids',
                    'names', 'identifier', 'identifiers', 'number', 'url')
BLACKLISTED_SHAPE_NAMES = ('ComponentChildList','FirewallManagerRuleGroups', 'Rules','HeaderNames', 'ExcludedRules', 'CountryCodes', 'CookieNames', 'TextTransformations', 'ComponentSummaryList','AnomalyMonitors','ConfigurationList','HandshakeResources','JsonPointerPaths','AdministrativeActions','DataValueList','ThemeValuesList','Expressions','CostCategoryRulesList')

console = get_rich_console()

def find_key_in_dict_keys(key:str, dict_keys: Union[list, dict]) -> str:
    """Case insensitive search for a `key` in a `list` or `dict.keys()`. 

    Args:
        key (str): Search case insensively for
        dict_keys (Union[list, dict]): In a list or keys of dictionary

    Returns:
        str: Found key that case insensively matches the given key.
    """
    dict_keys_list = dict_keys
    if type(dict_keys) == dict:
        dict_keys_list = dict_keys.keys()
    for dict_key in dict_keys_list:
        if key.lower() == dict_key.lower():
            return dict_key
    return False

def get_max_results_value_from_shape(input_shape: Shape) -> int:
    """Finds the `MaxResults` highest value for an input shape.

    Args:
        input_shape (Shape): Input shape of an operation. Generally postfixed with `Request`.

    Returns:
        int: `MaxResults` highest value if found in the Shape definition
    """
    flat_shapes_and_target_paths = flatten_shape_to_its_non_collection_shape_and_target_paths(input_shape)
    flat_members = [_.shape for _ in flat_shapes_and_target_paths]
    result = False
    for member_shape in flat_members:
        shape_key_name = getattr(member_shape, 'key_name', False)
        is_key_name_maxresults = shape_key_name and (shape_key_name.lower() == 'maxresults')
        shape_name = get_shape_name(member_shape)
        if shape_name.lower() == 'maxresults' or is_key_name_maxresults:
            member_shape
            max_value = member_shape.metadata.get('max', False)
            if max_value:
                return max_value
    return False

def get_input_shape(operation_model: OperationModel) -> Shape:
    """Get the input shape of the operation model

    Args:
        operation_model (OperationModel): botocore OperationModel 

    Returns:
        Shape: Input Shape of the OperationModel
    """
    return getattr(operation_model, 'input_shape', False)



def get_members_shapes(shape: Shape) -> List[Shape]:
    """Get the member shapes of and input or output Shape.

    Args:
        shape (Shape): Shape to get members from

    Returns:
        List[Shape]: Member Shapes, added `key_name` and `parent_name` values to objects.
    """
    found_members_shapes = []
    if not shape:
        return found_members_shapes
    if shape.name in BLACKLISTED_SHAPE_NAMES:
        return found_members_shapes
    if shape.type_name == 'structure':
        for shape_key, member in shape.members.items():
            setattr(member, 'key_name', shape_key)
            setattr(member, 'parent_name', shape.name)
            found_members_shapes.append(member)
    elif shape.type_name == 'list':
        only_member = shape.member
        found_members_shapes.append(only_member)
    elif shape.type_name == 'map':
        t = get_members_shapes(shape.value)
        return t
    return found_members_shapes


def filter_non_collection_shape(shape_and_target_path: ShapeAndTargetPath) -> bool:
    """Bool function to check ShapeAndTargetPath Named Tuple 

    Args:
        shape_and_target_path (ShapeAndTargetPath): Named Tuple. 

    Returns:
        bool: _description_
    """
    shape = shape_and_target_path.shape
    has_key_name = getattr(shape, 'key_name', False)
    is_non_collection = shape.type_name not in SHAPE_COLLECTION_TYPES
    return has_key_name and is_non_collection


def _flatten_shape_to_its_members_and_target_paths(shape, target_str='', ):
    result = []
    members = get_members_shapes(shape)
    for member in members:
        member_key_name = getattr(member, 'key_name', False)
        new_target_str = target_str
        if member_key_name:
            if target_str == '':
                new_target_str = f"{member_key_name}"
            else:
                new_target_str = f"{target_str}[*].{member_key_name}"
        if member.type_name in ('structure', 'list'):# TODO:SHAPE_COLLECTION_TYPES-('map',):
            inner_list = _flatten_shape_to_its_members_and_target_paths(member, new_target_str)
            result.extend(inner_list)
        else:
            result.append(ShapeAndTargetPath(member, new_target_str))
    return result

def flatten_shape_to_its_non_collection_shape_and_target_paths(shape):
    all_flat_members_and_target_paths = _flatten_shape_to_its_members_and_target_paths(shape)
    non_collection_shapes_and_target_paths = list(filter(filter_non_collection_shape, all_flat_members_and_target_paths))
    return non_collection_shapes_and_target_paths

"""
    root = Tree("🌲 [b green]Rich Tree", highlight=True, hide_root=True)
    node = root.add(":file_folder: Renderables", guide_style="red")
    simple_node = node.add(":file_folder: [bold yellow]Atomic", guide_style="uu green")
    simple_node.add(Group("📄 Syntax", syntax))
    simple_node.add(Group("📄 Markdown", Panel(markdown, border_style="green")))
    containers_node = node.add(
        ":file_folder: [bold magenta]Containers", guide_style="bold magenta"
    )
    containers_node.expanded = True
    panel = Panel.fit("Just a panel", border_style="red")
    containers_node.add(Group("📄 Panels", panel))

therewillbebeloodthere
        containers_node.add(Group("📄 [b magenta]Table", table))

    console = Console()

    console.print(root)
"""


def rich_str_shape(shape: Shape) -> str:
    """Transforms a Shape to rich supported string.

    Args:
        shape (Shape): botocore.model.Shape object.

    Returns:
        str: Rich string for shape.
    """
    key_name = getattr(shape, 'key_name', '{')
    type_name = str(shape.type_name)
    if key_name == '{' and type_name == 'structure':
        type_name = ''
    if type_name:
        type_name = f"({type_name})"
    shape_str = f"[blue]{key_name} [white]{type_name}"
    return shape_str


def generate_rich_tree_from_shape(shape: Shape) -> Tree:
    """Genereate a rich Tree containing `shape` and it's members."""
    tree = Tree(rich_str_shape(shape), guide_style="red")
    def _recursive_stringify_shape(shape, node):

        members = get_members_shapes(shape)
        for member in members:
            member_str = rich_str_shape(member) 
            if member.type_name in SHAPE_COLLECTION_TYPES:
                new_node = node.add(member_str)
                _recursive_stringify_shape(member, new_node)
            else:
                node.add(member_str)
    _recursive_stringify_shape(shape, tree)
    return tree
        
    
# @lru_cache(maxsize=500) # print(get_shape_name.cache_info())
def get_shape_name(shape: Shape) -> str:
    """Returns the shapes name, using custom set object values"""
    shape_name = shape.name
    shape_key_name = getattr(shape, 'key_name', False)
    if shape_name.lower() in UNWANTED_SHAPE_NAMES_LOWERED:
        # check for name forgotten in shape.serialization if the name is a type name.
        if shape_key_name:
            return shape_key_name
        elif shape.serialization and shape.serialization.get('name', False):
            given_serialization_name = shape.serialization.get('name')
            return given_serialization_name
    return shape_name
   
def get_required_parameter_shapes_from_operation_model(operation_model: OperationModel) -> List[Shape]:
    input_shape = get_input_shape(operation_model)
    if not input_shape:
        return [] # there's no input shape
    input_shape_members = get_members_shapes(input_shape)
    input_required_member_names = input_shape.required_members
    if input_required_member_names == []:
        return []
    
    required_member_shapes = []
    for input_member in input_shape_members:
        input_member_name = get_shape_name(input_member)
        for required_name in input_required_member_names:
            if compare_two_camel_case_words(required_name, input_member_name):
                required_member_shapes.append(input_member)
            
    return required_member_shapes

# def get_required_parameters_for_operation(operation_model):
#     required_parameter_shapes = get_required_parameters_for_operation(operation_model)
#     required_parameter_names = [
#         get_shape_name(r_param_shape)
#         for r_param_shape in required_parameter_shapes
#     ]
#     return required_parameter_names

def compare_shape_names(member, search_shape):
    return get_shape_name(member) == get_shape_name(search_shape)


